/*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <locale.h>
#include <string.h>
#include <ctype.h>

#ifdef ICONV
#include <iconv.h>
iconv_t ic;
#endif
/*
 224:  225:  226:  227:  228:  229:  230:  231:  232:  233:  234:  235:  236:  237:  238:  239:
 240:  241:  242:  243:  244:  245:  246:  247:  248:  249:  250:  251:  252:  253:  254:  255:
 */
static char keys[] = ".f,wltau[bqrkvyjgzhcne;dmspi'ox]";

void sendString(XKeyEvent *event, char *string);
void sendEvent(XKeyEvent *event);
void sendKey(XKeyEvent *event, int modifiers, int keycode);

int main(int argc, char **argv) {
    Display* display;
    int screen_num;
    char *display_name = getenv("DISPLAY");	
    Window window;   
    int revert_to, i;
    XKeyEvent event;
    
    
    if(argc < 2) {
        fprintf(stderr, "Usage: %s [ -n | -kCODE | -^X | -Fx ] \"string1\" ] ... \n", argv[0]);
        return -1;
    }

#ifdef ICONV
    char *lc = getenv("LC_ALL");
    if(NULL == lc || NULL == (lc = strchr(lc, '.'))) {
           fprintf(stderr, "%s: cannot determine your charset. Set LC_ALL variable!\n", argv[0]);
           return -1;
    } else {
        lc++;
        ic = iconv_open("koi8-r", lc);
        if (ic == (iconv_t) -1) {
           fprintf(stderr, "%s: Charset %s is not supported by your iconv library\n", argv[0], lc);
           return -1;
        }
    }
#endif

    display = XOpenDisplay(display_name);

    if(!setlocale(LC_CTYPE, "") || !XSupportsLocale()) {
         fprintf(stderr, "error: no locale support\n");
         return -1;
    }

    if (display == NULL) {
        fprintf(stderr, "%s: Cannot open display: %s\n", argv[0], display_name);
        return -1;
    }

    screen_num = DefaultScreen(display);
//    XSetInputFocus(display, PointerRoot, RevertToPointerRoot, CurrentTime);
    XSync(display, False);

    XGetInputFocus(display, &window, &revert_to);    // get window with focus
    if (window == None) {
        fprintf(stderr, "error: No window is focused\n");
        return 0;
    }
            
    event.display = display;
    event.window = window;
    event.root = RootWindow(event.display, DefaultScreen(event.display));
    event.subwindow = None;
    event.time = CurrentTime;

    event.x = event.y = event.x_root = event.y_root = 0;
    event.same_screen = True;
    
    for(i=1;i<argc;i++) {
        if(argv[i][0] == '-') {
            switch(argv[i][1]) {
                case 'n':   sendKey(&event, 0x10, 36); break;
                case '^':   sendKey(&event, 0x14, XKeysymToKeycode(event.display, argv[i][2])); break;
                case 'k':   sendKey(&event, 0x10, XKeysymToKeycode(event.display, atoi(argv[i] + 2))); break;
                case 'f':
                case 'F':   sendKey(&event, 0x10,  argv[i][2]+18); break;
                default:
                    fprintf(stderr, "%s: unknown key `%s'\n", argv[0], argv[i]+1);
            }
        } else sendString(&event, argv[i]);
    }

#ifdef ICONV   
    iconv_close(ic);
#endif
    
    XSync(display, False);
    XFlush(display);    
    XCloseDisplay(display);
    return 0;
}

void sendKey(XKeyEvent *event, int modifiers, int keycode) {
    event->state = modifiers;
    event->keycode = keycode;
    sendEvent(event);
}

void sendString(XKeyEvent *event, char *str) {
    int i, l = strlen(str);
    char c;
#ifdef ICONV
    size_t f = l, t = l;
    char *in, *b;
    in  = str;
    b = str = alloca(l);
    iconv(ic, &in, &f, &str, &t);
    l -= t;
    b[l] = 0;
    str = b;
#endif

    for(i = 0; i < l; i++) {
        c = str[i];
        event->state = 0;

        if(c < 0) {
            event->state += 0x2000;

            if((c & 0x20) > 0) {
                c ^= 0x20;
                event->state ++;
            }
            c = keys[64 + c];

        } else if((c & 0x20) == 0) {
            event->state = 1;
            printf("%c\n", c);
            c |= 0x20;
        } else if((c > 32 && c < 42) || c == 58)
            event->state = 1;

        event->keycode = XKeysymToKeycode(event->display, c);
        sendEvent(event);
    }
}


void sendEvent(XKeyEvent *event) {
        event->type = KeyPress;
        XSendEvent(event->display, event->window, True, KeyPressMask, (XEvent *) event);

        event->type = KeyRelease;
        XSendEvent(event->display, event->window, True, KeyPressMask, (XEvent *) event);
}
